#include <l4/types.h>
#include <l4/ipc.h>
#include <l4/syscalls.h>

#include <thread.h>


Thread::Thread (unsigned nr)
{
  m_myself=l4_myself();
  m_myself.id.lthread=nr;
}

Thread::Thread (l4_threadid_t id) 
{
  m_myself=id;
}

int Thread::Create(void *ptrStack, unsigned char prio) 
{
  qword_t dummy;
  qword_t *Stack = (qword_t *)ptrStack;
  l4_threadid_t pager, preempter;

  if (l4_myself().id.task != m_myself.id.task) return false;

  m_prio = prio;
  pager = preempter = L4_INVALID_ID;

  /* Get Pager & Preempter */
  l4_thread_ex_regs (l4_myself().id.lthread, (qword_t)-1, (qword_t)-1, 
		     &preempter, &pager, &dummy, &dummy);

  /* Start Thread */
  *Stack-- = (qword_t)this;
  l4_thread_ex_regs (m_myself.id.lthread, (qword_t)__asm_thread_start, (qword_t)(ptrStack),
		     &preempter, &pager, &dummy, &dummy);
  return true;
}


Thread::~Thread (void)
{
  qword_t dummy;
  l4_threadid_t pager, preempter;

  pager = preempter = L4_INVALID_ID;
  l4_thread_ex_regs (m_myself.id.lthread, (qword_t)Thread::ThreadStop, (qword_t)-1,
		     &preempter, &pager, &dummy, &dummy);
  
}

int Thread::SetPriority (unsigned char new_prio)
{
  l4_schedule (m_myself, L4_NIL_ID, new_prio);
  return 0;
}

__asm__  (
	  "__asm_thread_start:				\n\t"
	  "ldgp $29,0($27)				\n\t"
	  "ldq $16, 0($30)				\n\t"
	  "subq $30, 8, $30				\n\t"
	  "lda $27, ThreadStart__6ThreadP6Thread	\n\t"
	  "jsr $26, ($27), ThreadStart__6ThreadP6Thread	\n\t"
	  "1: br 1b;"
	  );


void Thread::ThreadStart(Thread *thread) 
{
  thread->m_myself = l4_myself ();
  if (thread->m_prio)
    thread->SetPriority (thread->m_prio);
  thread->ThreadFunction();
  ThreadStop();
}

void Thread::ThreadStop() 
{
  l4_alpha_ipc_receive(l4_myself(), NULL, NULL, L4_IPC_NEVER);
}


l4_msgdope_t Thread::SendShort(l4_timeout_t timeout) {
  return l4_alpha_ipc_send(m_myself, NULL, NULL, timeout);
}

l4_msgdope_t Thread::SendShort(l4_timeout_t timeout, l4_ipcregs_t &regs) {
  return l4_alpha_ipc_send(m_myself, NULL, &regs, timeout);
}

l4_msgdope_t Thread::ReceiveShort(l4_timeout_t timeout, l4_ipcregs_t &regs) {
  return l4_alpha_ipc_receive(m_myself, NULL, &regs, timeout);
}

l4_msgdope_t Thread::ReceiveShort(l4_timeout_t timeout) {
  return l4_alpha_ipc_receive(m_myself, NULL, NULL, timeout);
}
  
l4_msgdope_t Thread::Call(const void *snd_msg, const l4_ipcregs_t *snd_regs,
		    void *rcv_msg,  l4_ipcregs_t *rcv_regs, l4_timeout_t timeout) {
  return l4_alpha_ipc_call(m_myself, snd_msg, snd_regs, rcv_msg, rcv_regs, timeout);
} 

l4_msgdope_t Thread::WaitShort(l4_timeout_t timeout, l4_threadid_t *src, l4_ipcregs_t *rcv_regs) {
  return l4_alpha_ipc_wait(src, NULL, rcv_regs, timeout);
}



